@@ -15,7 +15,7 @@
 ##    You should have received a copy of the GNU General Public License
 ##    along with Fonty Python.  If not, see <http://www.gnu.org/licenses/>.
 
-import os, sys, locale, codecs, glob
+import os, sys, locale, glob
 import Image, ImageFont, ImageDraw 
 import fontybugs, fpsys
 from pathcontrol import *
@@ -29,9 +29,26 @@ class FontItem( object ):
     Never instantiate one directly.
     """
     def __init__(self, glyphpaf ):
+        ## glyphpaf must be whatever it is. unicode or a byte string
+        ## This is used to access files themselves.
+        ## When the LANG encoding is utf8, its Unicode.
+
         self.glyphpaf = glyphpaf
+        #print "FontItem init, glyphpaf:", [glyphpaf]
+        ## I want to have a var I can use when I display the glyphpaf
+        ## either in the gui or onto the cli. This should be a unicode
+        ## object, so I must make sure of it's type before I do so.
+        if type(glyphpaf) is str:
+            ## It's a byte string, not unicode. Convert it.
+            self.glyphpaf_unicode = glyphpaf.decode(locale.getpreferredencoding(), "replace")
+        else:
+            self.glyphpaf_unicode = glyphpaf
+
+        ## The same goes for name. It *must* be unicode.
         self.name = os.path.basename ( glyphpaf )
-        
+        if type(self.name) is str:
+            self.name = self.name.decode(locale.getpreferredencoding(), "replace")
+            
         self.ticked = False # State of the tick/cross symbol.
         self.inactive = False # Set in fpsys.markInactive()
         self.msg = "" #Say something unique when I draw this item.  
@@ -81,10 +98,10 @@ class FontItem( object ):
         
         i = 0
         ## Step through all subfaces.
-        #print "BUILDING:", self.glyphpaf
+        #print "BUILDING:", [self.glyphpaf]
         while True:
             try:
-                fileDoesExist = os.path.exists(self.glyphpaf)
+                fileDoesExist = os.path.exists( self.glyphpaf )
                 if not fileDoesExist:
                     fpsys.logBadStrings( self.glyphpaf )
                     self.badfont = True
@@ -104,39 +121,12 @@ class FontItem( object ):
                 NOTE: Sets badfont
                 """
                 if i == 0: # fubar on the first face:
-                    #print "IOERROR:", self.glyphpaf
-                    fpsys.logBadStrings( self.glyphpaf )
                     self.badfont = True
                     self.badfontmsg = _("Font may be bad and it cannot be drawn.")
                     self.badstyle = "PIL_IO_ERROR"
                     ## If the multi face font is damaged after the
                     ## first face, then this won't catch it...
                 break # it's at the end of the sub-faces
-                if False: # using this to remark all the remarks. It collapses in my editor.
-                    pass
-                    ##            ## I did not want to do this next part, it *must* slow things down.
-                    ##            ## I found a font throwing a MemoryError (Onsoku Seinen Plane.ttf)
-                    ##            ## that only happens upon the .text() command.
-                    ##            ## So, to know if this will happen to others, I am forced to render
-                    ##            ## a bit of text here and now. :(
-                    ##            ##
-                    ##            ## UPDATE: It's not working. Onsoku *only* barfs on "TE" and not
-                    ##            ## "A" or even chr(0) to chr(255) all in a string...
-                    ##            ## So, it's virtually impossible to know at this point what will
-                    ##            ## cause the MemoryError in the rendering step.
-                    ##            ## So, I am remarking this stuff and will invent a new flag called
-                    ##            ## specialvoodooerror which will be handled as a special case in the
-                    ##            ## Fitmap drawing routines in the gui.
-                    ##            try:
-                    ##                print "RENDERING A"
-                    ##                text = "A"
-                    ##                pilimage = Image.new("RGB", (font.getsize( text )), (255,255,255)) 
-                    ##                drawnFont = ImageDraw.Draw( pilimage ) # Draws INTO pilimage 
-                    ##                drawnFont.text((0,0) , text, font=font, fill=0 ) # Causes MemoryError
-                    ##            except MemoryError:
-                    ##                print "CAUGHT IT FOR:", self.glyphpaf
-                    ##                fpsys.logBadStrings( self.glyphpaf )
-                    ##                self.badfont = True
             except UnicodeEncodeError:
                 """
                 NOTE: Sets badfont
@@ -149,27 +139,35 @@ class FontItem( object ):
                 fpsys.logBadStrings( self.glyphpaf )
                 self.badfont = True
                 self.badfontmsg = _("Font may be bad and it cannot be drawn.")
-                #self.badstyle = "BAD_UNICODE"
-                ## Changed to BAD_OPEN because it seems that if the glyphpaf
-                ## is bad unicode, then this error fires before the IOError
-                ## which means that this font may also not be on that path
-                ## at all (i.e. BAD_OPEN) and I don't want to let users 
-                ## work with bad fonts like this, so I am foreclosing the
-                ## bad unicode distinction.
                 self.badstyle = "PIL_UNICODE_ERROR"
                 break
 
             except:
                 ## What next on the error pile?
+                print "CORNER CASE in FontItem.__queryFontFamilyStyleFlagBad:", [self.glyphpaf]
+                print sys.exc_info()
                 raise
                 ## Abort the app because this is an unhandled PIL error of some kind.
             if not self.badfont:
-                ## No point Try-ing here, this segfaults when style/family is Null.
-                fpsys.logSegfaulters( self.glyphpaf ) # record the potential destroyer of apps!
-                self.family.append( font.getname()[0] )
-                self.style.append( font.getname()[1] )
+                ## *If* 'check' is run, there will be a file containing pafs of the
+                ## fonts that segfault PIL. (To my best knowledge, at least.)
+                ## This file is held in the 'segfonts' list - opened in fpsys
                 
-                i += 1
+                ## So, before we do a getname and cause a segfault, let's
+                ## see whether that font is in segfonts, and if so, skip it.
+                if self.glyphpaf not in fpsys.segfonts:
+                    ## No point Try-ing here, this segfaults when style/family is Null.
+                    fpsys.logSegfaulters( self.glyphpaf ) # record the potential destroyer of apps!
+                    self.family.append( font.getname()[0] )
+                    self.style.append( font.getname()[1] )
+                    i += 1
+                else:
+                    ## It WAS in the list! So, we can flag it and get on with life :)
+                    #print "SKIPPING:",[self.glyphpaf]
+                    self.badfont = True
+                    self.badfontmsg = _("Font causes a segfault. Fix or remove it.")
+                    self.badstyle = "PIL_SEGFAULT_ERROR"
+                    break
                 
         self.numFaces = i
                 
@@ -187,15 +185,17 @@ class FontItem( object ):
         be seen.
         
         """
-        paf, points, text = self.glyphpaf, fpsys.config.points, " " + fpsys.config.text + " "
+        ## text gets extra spaces at the end to cater for cut-off characters.
+        paf, points, text = self.glyphpaf, fpsys.config.points, " " + fpsys.config.text + "  "
         i = 0
         while (True):
             try:
                 font = ImageFont.truetype(paf, points,index=i, encoding=enc) 
-
-                pilheight = int( font.getsize( fpsys.config.text )[1] )# Must int for PIL.
+                w,h = font.getsize( text )
+                pilheight = int(h)
+                pilwidth = int(w)
                 pilheight = min( pilheight, maxHeight )
-                pilimage = Image.new("RGB", (width,pilheight), (255,255,255)) 
+                pilimage = Image.new("RGB", (pilwidth, pilheight), (255,255,255)) 
                 
                 if self.inactive:
                     col = (230, 230, 230) 
@@ -212,7 +212,7 @@ class FontItem( object ):
                 ## All is well, so we step ahead to the next *potential* sub-face
                 ## and return the font image data.
                 i += 1
-                yield pilimage, pilheight  
+                yield pilimage, pilheight, pilwidth
 
             except MemoryError:
                 """
@@ -220,9 +220,16 @@ class FontItem( object ):
 
                 This one CAN ONLY BE CAUGHT HERE.
                 **IDEALLY**, it should have been caught in __queryFontFamilyStyleFlagBad
-                but for reasons explained there, it cannot.
+                but for reasons explained below, it cannot.
                 So, we have a badfont flag being set here too :(
                 """
+                ## I found a font throwing a MemoryError (Onsoku Seinen Plane.ttf)
+                ## that only happens upon the .text() command.
+                ##
+                ## UPDATE: Clever tricks don't work. Onsoku *only* barfs on "TE" and not
+                ## "A" or even chr(0) to chr(255) all in a string...
+                ## So, it's virtually impossible to know at this point what will
+                ## cause the MemoryError in the rendering step.
                 self.badfontmsg = _("Font cannot be drawn.")
                 self.badstyle = "PIL_CANNOT_RENDER"
                 fpsys.logBadStrings( self.glyphpaf )
@@ -289,7 +296,17 @@ def itemGenerator( fromObj, sourceList ):
     instantiated according to the type of the font.
     Call it once-off; or in a loop to get one FontItem after another.
     Pass it a sourceList that contains pafs.
+    
+    VERY NB:
+    When the app is run from a utf8 locale, this func generates 
+    UNICODE glyphpaf vars.
+    When run from C/POSIX/None it generates BYTE STRING glyphpaf vars.
+    
+    sourceList is not a predictable beast. It comes in mixed strings/unicode
     """
+    #print "sourceList comes in:",[sourceList]
+    #print
+    
     def ext(s): return s.split(".")[-1].upper()
     def stripExt(s): return s[:s.rfind(".")]
     
@@ -392,9 +409,11 @@ def itemGenerator( fromObj, sourceList ):
         for paf in TTCList:
             fi = TruetypeCollectionItem( paf )
             listOfItemsGenerated.append(fi)
-        
-    ## Sort the list: not sure this will work....
-    listOfItemsGenerated.sort( cmp=locale.strcoll, key=lambda obj:obj.glyphpaf ) # Try to sort on that field.
+    
+    ## NB: listOfItemsGenerated can contain MIXED byte strings/unicode
+    
+    ## Sort the list: I use the glyphpaf_unicode var becuase it's unicode only.
+    listOfItemsGenerated.sort( cmp=locale.strcoll, key=lambda obj:obj.glyphpaf_unicode ) # Try to sort on that field.
 
     ## Supply it: This is pure magic!
     for fi in listOfItemsGenerated:
@@ -438,12 +457,20 @@ class Folder(BasicFontList):
     """
     Represents an entire Folder (from a path given by user clicking on the
     GenericDirCtrl or from a commanline string.)
+    
+    This is called from fpsys.instantiateViewFolder
+    
     Contains a list of various FontItem Objects.
     Supply the path
     """
     def __init__(self, path):
         BasicFontList.__init__(self)
-        ## Public var
+        #print "path:",[path]
+        
+        ## I still have to verify this, but:
+        ## I reckon path is always coming in as unicode.
+        ## From the gui, it's unicode anyway cos of the dir control.
+        ## From the cli, I converted args to unicode there.
         self.path = os.path.abspath(path) # fix relative paths
 
         ## NEW INFO : DEC 2007
@@ -456,23 +483,22 @@ class Folder(BasicFontList):
         ##  ("locale") for the proper coding."
         ##
         ## On Linux: os.path.supports_unicode_filename is always == True
-        ## On my system:
+        ## On my system, with LANG=en_ZA.utf8
         ## >>> locale.getpreferredencoding()
         ## 'UTF-8'
         ##    Return the charset that the user is likely using,
         ##    according to the system configuration.
+        ## With LANG=C it returns "ANSI****"
         ## On my system:
         ## >>> sys.getfilesystemencoding()
         ## 'UTF-8'
         ## This one returns the ENCODING (byte string to unicode) needed to
         ## convert filenames from the O/S *to* unicode.
-        ##
         
         listOfFilenamesOnly = []
         ## If self.path is unicode, return will be list of unicode objects (with some slip-ups...)
-        ## I have forced the app to be 100% unicode, so this is going to be a Unicode list.
         ## self.path comes from the GenericDirCtrl, which is part of a Unicode wxPython build
-        ## therefore it always emits Unicode objects. (Or it comes from cli)
+        ## therefore it always emits Unicode objects. (Or it comes from cli, which also converted)
         listOfFilenamesOnly = os.listdir (  self.path  ) # Get the unicode list
         
         ## I could use this:
@@ -486,20 +512,30 @@ class Folder(BasicFontList):
                 sourceList.append( paf )
             except:
                 ## Okay, 'f' was not able to join to self.path
-                ## I'ts prob a UnicodeDecodeError, but I don't care.
-                ## This is one font that will never make it through
-                ## to the app. It won't even become a 'badfont'
-                fpsys.logBadStrings(f) # I'm keeping a list of bad names.
+                ## This is because 'f' is a byte string, not a unicode
+                ## That means that os.listdir couldn't figure it out...
+                ## I will revert to byte strings and try again:
+                #print "Could not join to path name:"
+                #print [self.path], " + ", [f]
+                try:
+                    paf = os.path.join( self.path.encode( locale.getpreferredencoding()), f )
+                    sourceList.append( paf )
+                except:
+                    ## No, could not get a hold of that file.
+                    ## In theory, this should never happen.
+                    print "CORNER CASE in Folder.__init__:", [self.path], " + ", [f]
+                    fpsys.logBadStrings(f) # I'm keeping a list of bad names.
 
         ## Now employ the generator magic:
+        ## Makes FontItem objects for each paf in the list.
         for fi in itemGenerator( self, sourceList ):
+            #print "fi:",[fi]
             self.append(fi)
         
         if len(self) == 0:
+            print "EMPTY FOLDER"
             raise fontybugs.FolderHasNoFonts(self.path)
-          
-        ## I sorted the list in the generator, so if that didn't work, perhaps this will
-        ## self.sort ( locale.strcoll ) # Sort myself....
+
     def __str__(self):
         return str(self.path)
         
@@ -522,23 +558,36 @@ class Pog(BasicFontList):
     """
     def __init__(self, name ): #, progressCallback = None):
         BasicFontList.__init__(self)
-        ##private props
-        #self.__progressCallback = progressCallback
-        self.__pc = PathControl() # My own path control.
+        self.__pc = fpsys.iPC # A hack to pathcontrol.
         
         ## Public properties:
-        self.name = name
+        ##
+        ## name always comes in as a byte string because
+        ## we built the path up to .fontypython from byte strings
+        
+        #print "Pog create"
+        #print "name coming in as:", [name] # So far, for all locales, this is always str.
+        ## Make a unicode of that name:
+        if type(name) is str:
+            uname = name.decode(locale.getpreferredencoding(),"replace")
+        else:
+            uname = name
+        ## Stores a unicode for access from other places:
+        self.name = uname
+        
         self.__installed = "dirty" #am I installed?
-        #self.empty = None #do I have fonts in me?
-        self.paf = self.__pc.appPath() + self.name + ".pog"
-        #print "Pog __init__ says self.paf is of type:", type(self.paf)
+        
+        ## Note, name (not self.name) is used here. 
+        ## We are continuing to use pure byte strings for the paf
+        self.paf = os.path.join( self.__pc.appPath(),name + ".pog")
+
         self.badpog = False #To be used mainly to draw icons and ask user to purge.
         
     def label(self):
         """
-        A handy way to refer to Folders to Pogs in certain circumstances.
+        A handy way to refer to Folders & Pogs in certain circumstances.
         See around line 1296 wxgui.OnMainClick()
-        Pog.label returns the name
+        Pog.label returns the name (in unicode)
         Folder.label return the path
         These are both the full paf of the font.
         """        
@@ -551,19 +600,33 @@ class Pog(BasicFontList):
         ## NB: NO error is raised if an empty file is opened and read...
         ## If there is some chronic hard drive problem, I assume Python will quit anyway...
         try:
-            f = codecs.open( self.paf, "r", sys.getfilesystemencoding())
+            #print "Trying to open:", [self.paf]
+            ## In theory, any file can *always* be opened - no matter what
+            ## locale. POSIX deals only in byte strings. So, this next
+            ## line will never fail.
+            
+            ## I am going to open it as an ASCII file:
+            f = open( self.paf, 'r' ) # ASCII byte string file only.
         except:
-            raise fontybugs.PogInvalid( self.paf )
+            print "CORNER CASE in __openfile on paf:", [self.paf]
+            print sys.exc_info()
+            raise SystemExit
+
         ## Let's see what kind of line 1 we have
         line1 = f.readline()[:-1]
+
         self.__installed = "dirty" # unsure as to the status
+
         if line1.upper() == "INSTALLED": self.__installed = "yes"
+
         if line1.upper() == "NOT INSTALLED": self.__installed = "no"
+
         if self.__installed == "dirty":
             ## We have a bad pog.
-            print "ABOUT TO RENAME POG"
+            #print "ABOUT TO RENAME POG"
             self.__renameBadPog()
             raise fontybugs.PogInvalid( self.paf )
+
         ## At this point, we have a valid pog file:
         ## It has a valid line 1
         ## It may or may not have paf lines below it.
@@ -572,7 +635,7 @@ class Pog(BasicFontList):
         
     def isInstalled(self):
         """
-        Will ONLY raise a PogInvalid error. Any other will abort app.
+        Passes a raise PogInvalid error through. Any other will abort app.
         """
         if self.__installed == "yes": return True
         if self.__installed == "no": return False
@@ -605,23 +668,34 @@ class Pog(BasicFontList):
         self.clear() #clear is in basicfontlist.py
 
         sourceList = []
-        for paf in f: #This continues from line 2 onwards ...
-            paf = paf[:-1] #Strip the damn \n from the file
-            sourceList.append(paf) #[path,filename])
-        f.close()  
-        
+        ## Right,
+        ## If a line of the CONTENT of f is encoded differently to what the locale is
+        ## then the for paf in f: line throws a UnicodeDecodeError
+        ## This means that paf can't be read, but perhaps others can be...
+        try:
+            for paf in f: #This continues from line 2 onwards ...
+                paf = paf[:-1] #Strip the damn \n from the file
+                sourceList.append(paf)
+            f.close()  
+        except UnicodeDecodeError:
+            ## I can't even display the paf because it's not been set
+            ## (paf is the last value that was read, since this is an error condition)
+            ## I don't think I have a choice here but to simply pass
+            pass
+            ## Not using this anymore.
+            #raise fontybugs.PogContentEncodingNotMatched( self.paf )
+            
         ## Now to make Fontitems out of sourceList
         for fi in itemGenerator( self, sourceList):
             self.append(fi) # store them in myself.
             
     def purge(self):
         """
-        ####
-        ## Purge method - remove fonts in the pog that are not on disk.
-        ##
-        ## Raises
-        ##          PogEmpty
-        ##          PogInstalled
+        Purge method - remove fonts in the pog that are not on disk.
+
+        Raises
+                 PogEmpty
+                 PogInstalled
         """
         ## can't purge an empty pog
         if len(self) == 0:
@@ -637,7 +711,7 @@ class Pog(BasicFontList):
                     if not os.path.exists(i.glyphpaf) :
                         badfonts.append(i) 
                 except:
-                    pass # it's a bad through and through! It'll be axed too.
+                    pass # it's bad through-and-through! It'll be axed too.
             ## Now go thru this list and remove the bad items.
             for bi in badfonts:
                 #print "purging:", bi.name
@@ -646,24 +720,22 @@ class Pog(BasicFontList):
 
     def install(self):
         """
-        ####
-        ## Install the fonts in myself to the user's fonts folder.
-        ## NOTE:
-        ## Even if ONLY ONE font out of a gigazillion in the pog 
-        ## actually installs, the POG == INSTALLED.
-        ## If we have a font that cannot be sourced, flag BADPOG
-        ##
-        ## For Type1 fonts - 
-        ## The choice I have made is:
-        ## I will ONLY reference the PFB in the Pog file.
-        ## When we install a Pog, the AFM/PFM must be sought
-        ## in the original folder and linked.
-        ## For uninstall, simply kill the associated AFM/PFM
-        ##
-        ## Raises:
-        ##          PogEmpty
-        ##          PogAllFontsFailedToInstall
-        ##          PogSomeFontsDidNotInstall
+        Install the fonts in myself to the user's fonts folder.
+        NOTE:
+        Even if ONLY ONE font out of a gigazillion in the pog 
+        actually installs, the POG == INSTALLED.
+        If we have a font that cannot be sourced, flag BADPOG
+
+        For Type1 fonts - 
+            The choice I have made is:
+            I will ONLY reference the PFA/B in the Pog file.
+            When we install a Pog, the metric file must be sought
+            in the original folder and linked.
+
+        Raises:
+            PogEmpty
+            PogAllFontsFailedToInstall
+            PogSomeFontsDidNotInstall
         """        
         ## We start thinking all is rosey:
         self.__installed = "yes"
@@ -675,15 +747,16 @@ class Pog(BasicFontList):
         ## Now we go through the guts of the pog, font by font:
         bugs = 0
         for fi in self:
-            linkDestination = os.path.join(self.__pc.userFontPath(), fi.name) 
             
+            dirname = os.path.basename( fi.glyphpaf )
+            linkDestination = os.path.join(self.__pc.userFontPath(), dirname )
+
             ## Link it if it ain't already there.
             ## I am not checking for errors on os.symlink ...
-            
             if not os.path.exists(linkDestination):  
                 if os.path.exists(fi.glyphpaf):
                     os.symlink(fi.glyphpaf, linkDestination)  #Should do the trick.
-                    ## Now, the Type1 special case, link the AFM/PFM
+                    ## Now, the Type1 special case, link the metric file.
                     if isinstance( fi, Type1Item ):
                         ## It's a Type 1, does it have a metricpaf?
                         if fi.metricpaf:
@@ -705,24 +778,24 @@ class Pog(BasicFontList):
 
     def uninstall(self):
         """
-        ####
-        ## Uninstall the fonts.
-        ## NOTE:
-        ## If any font is NOT removed POG = INSTALLED
-        ## Any links that are not found = just ignore: They could have been removed by another pog, or this
-        ## could have been a bad pog anyway.
-        ## NO BAD POG flag EVER.
-        ##
-        ## Raises:
-        ##          PogEmpty
-        ##          PogLinksRemain
-        ##          PogNotInstalled
+        Uninstall the fonts.
+        NOTE:
+        If any font is NOT removed POG = INSTALLED
+        Any links that are not found = just ignore: They could have been removed by another pog, or this
+        could have been a bad pog anyway.
+        NO BAD POG flag EVER.
+
+        Raises:
+                 PogEmpty
+                 PogLinksRemain
+                 PogNotInstalled
         """        
         if len(self) == 0: raise fontybugs.PogEmpty # RAISED :: PogEmpty
         bugs = 0
         if self.__installed == "yes":
             for fi in self:
-                link = os.path.join(self.__pc.userFontPath(), fi.name) #Point at the link in .fonts folder.
+                dirname = os.path.basename( fi.glyphpaf )
+                link = os.path.join(self.__pc.userFontPath(), dirname )
                 ## Step one - look for the actual file (link)
                 if os.path.exists(link):
                     try:
@@ -754,27 +827,31 @@ class Pog(BasicFontList):
 
     def write(self) :
         """
-        ####
-        ## Write a pog to disk.
+        Write a pog to disk.
         """
         try:
-            #f = open(self.paf, "w") 
-            f = codecs.open( self.paf, 'w', sys.getfilesystemencoding() )
+            f = open( self.paf, 'w' ) # Going to make the contents ASCII only. Byte strings.
             i = "not installed\n"
             if self.__installed == "yes":
                 i = "installed\n"
             f.write(i) 
             #Now write the font pafs
             for i in self:
-                f.write(i.glyphpaf + "\n") 
+                ## since the glyphpaf can vary it's type
+                ## we must encode it to a byte string if it's unicode.
+                if type(i.glyphpaf) is unicode:
+                    gpaf = i.glyphpaf.encode( locale.getpreferredencoding() )
+                else:
+                    gpaf = i.glyphpaf
+                f.write( gpaf + "\n") 
             f.close() 
         except:
+            raise
             raise fontybugs.PogWriteError(self.paf)
 
     def delete(self):
         """
-        ####
-        ## Delete my pogfile, then clean myself up, ready to be destroyed.        
+        Delete my pogfile, then clean myself up, ready to be destroyed.        
         """
         try:
             os.unlink(self.paf)
